@@ -1,2 +1,5 @@
+"""
+Input/output methods declaration.
+"""
 from .inputparser import ReadYaml
 from .outmesh import WriteMesh
