@@ -6,23 +6,34 @@ import glob
 import shutil
 import petsc4py
 import numpy as np
+from scipy import spatial
+from scipy import ndimage
 
 from mpi4py import MPI
-from petsc4py import PETSc
 from time import process_time
 
 petsc4py.init(sys.argv)
-MPIrank = PETSc.COMM_WORLD.Get_rank()
-MPIsize = PETSc.COMM_WORLD.Get_size()
+MPIrank = petsc4py.PETSc.COMM_WORLD.Get_rank()
+MPIsize = petsc4py.PETSc.COMM_WORLD.Get_size()
 MPIcomm = MPI.COMM_WORLD
 
 
 class WriteMesh(object):
     """
-    Output model paramters using hdf5 library
+    Class for writing model outputs. The outputs are written as hdf5 files for each
+    mesh partition. The outputs structure is composed of library
+
+    .. note::
+
+        The model outputs are all located in an output folder (`dir` key in the inputfile documentation) and consist of a time series file named `gospl.xdmf` and 2 other folders (`h5` and `xmf`).
+
+    The `XDMF` file is the main entry point for visualising the output and should be sufficient for most users. This file can easely be opened within `Paraview <https://www.paraview.org/download/>`_.
     """
 
     def __init__(self):
+        """
+        Initialise model outputs paramters.
+        """
 
         self.step = 0
         self.stratStep = 0
@@ -33,6 +44,8 @@ class WriteMesh(object):
         # Sync the chosen output dir to all processors
         self.outputDir = MPIcomm.bcast(self.outputDir, root=0)
 
+        # In case of a restarting simulation, get the corresponding time
+        # values according to the restarting step
         if self.rStep > 0:
             self.step = self.rStep + 1
             if self.strat > 0:
@@ -41,15 +54,51 @@ class WriteMesh(object):
             self.tNow = self.tStart + self.rStep * self.tout
             self.rStart = self.tStart + self.rStep * self.tout
             self.saveTime = self.tNow + self.tout
-            self.saveStrat = self.tNow + self.strat
+            self.saveStrat = self.tEnd + self.tout
+            if self.strat > 0:
+                self.saveStrat = self.tNow + self.strat
         else:
             self.rStart = self.tStart
 
+        if self.strataFile is not None:
+            self.stratStep = self.initLay
+
+        # If series of paleomaps are used to force the model through time.
+        if self.forceStep >= 0:
+
+            res = 0.5
+            nx = int(360.0 / res)
+            ny = int(180.0 / res)
+            lon = np.linspace(0.0, 360.0, nx)
+            lat = np.linspace(0, 180, ny)
+
+            lon, lat = np.meshgrid(lon, lat)
+            latlonreg = np.dstack([lat.flatten(), lon.flatten()])[0]
+            self.regshape = lon.shape
+            outTree = spatial.cKDTree(self.lLatLon, leafsize=10)
+            distances, self.forceIDs = outTree.query(latlonreg, k=3)
+            self.forceWeights = 1.0 / distances ** 2
+            self.forceTop = 1.0 / np.sum(self.forceWeights, axis=1)
+            self.forceonIDs = np.where(distances[:, 0] == 0)[0]
+
+            outTree = spatial.cKDTree(latlonreg, leafsize=10)
+            distances, self.regIDs = outTree.query(self.lLatLon, k=3)
+            self.regWeights = 1.0 / distances ** 2
+            self.regTop = 1.0 / np.sum(self.regWeights, axis=1)
+            self.regonIDs = np.where(distances[:, 0] == 0)[0]
+
+            del distances, outTree, lon, lat, latlonreg
+            gc.collect()
+
+        # Petsc vectors
+        self.upsG = self.hGlobal.duplicate()
+        self.upsL = self.hLocal.duplicate()
+
         return
 
     def visModel(self):
         """
-        Visualise model outputs.
+        Main function to write model outputs on disk.
         """
 
         # Output time step for first step
@@ -57,7 +106,7 @@ class WriteMesh(object):
             self._outputMesh()
             self.saveTime += self.tout
 
-        # Output time step
+        # Output time step after start time
         elif self.tNow >= self.saveTime:
             self._outputMesh()
             self.saveTime += self.tout
@@ -71,8 +120,6 @@ class WriteMesh(object):
                 self.steppaleo = 0
                 self.newForcing = False
                 self.forceStep += 1
-                if self.forceStep >= self.forceNb:
-                    self.forceStep = 0
             else:
                 self.newForcing = True
 
@@ -80,7 +127,13 @@ class WriteMesh(object):
 
     def _createOutputDir(self):
         """
-        Create a directory to store outputs.
+        Create a directory to store outputs. By default the folder will be called
+        `output`. If a folder name is specified in the YAML input file, this name
+        will be used.
+
+        .. note::
+            The input option `makedir` gives the ability to delete any existing output folder with the same name (if set to `False`) or to create a new folder with the given dir name plus a number at the end (*e.g.* `outputDir_XX` if set to `True` with `XX` the run number). It prevents overwriting on top of previous runs.
+
         """
 
         # Get output directory
@@ -105,40 +158,110 @@ class WriteMesh(object):
 
         return
 
-    def outputStrat(self):
+    def _outputStrat(self):
         """
         Saves mesh local stratigraphic information stored in the DMPlex
-        to HDF5 file.
+        as HDF5 file. The following variables will be recorded:
+
+        - elevation at time of deposition, considered to be to the current elevation
+          for the top stratigraphic layer `stratZ`.
+        - thickness of each stratigrapic layer `stratH` accounting for both
+          erosion & deposition events.
+        - proportion of fine sediment `stratF` contains in each stratigraphic layer.
+        - porosity of coarse sediment `phiS` in each stratigraphic layer computed at
+          center of each layer.
+        - porosity of fine sediment `phiF` in each stratigraphic layer computed at
+          center of each layer.
+        - proportion of carbonate sediment `stratC` contains in each stratigraphic layer
+          if the carbonate module is turned on.
+        - porosity of carbonate sediment `phiC` in each stratigraphic layer computed at
+          center of each layer when the carbonate module is turned on.
+
+        .. important::
+
+            It is worth mentioning that the stratigraphic architecture is only outputed as
+            HDF5 files and does not record the XMF and XDMF files. A set of post-processing
+            scripts are then required to extract the informations and visualise the stratigraphic
+            records of any specific simulations.
         """
 
         t = process_time()
         h5file = (
             self.outputDir
             + "/h5/stratal."
-            + str(self.stratStep)
+            + str(self.step)
             + ".p"
             + str(MPIrank)
             + ".h5"
         )
         with h5py.File(h5file, "w") as f:
+
+            # Write stratal layers elevations per layers
             f.create_dataset(
-                "elev",
-                shape=(len(self.lcoords[:, 0]), 1),
+                "stratZ",
+                shape=(self.npoints, self.stratStep),
                 dtype="float32",
                 compression="gzip",
             )
-            f["elev"][:, 0] = self.hLocal.getArray()
-            f.create_dataset("sea", shape=(1, 1), dtype="float32", compression="gzip")
-            f["sea"][:, 0] = self.sealevel
+            f["stratZ"][:, : self.stratStep] = self.stratZ[:, : self.stratStep]
+
+            # Write stratal layers thicknesses per layers
             f.create_dataset(
-                "erodep",
-                shape=(len(self.lcoords[:, 0]), 1),
-                dtype="float32",
+                "stratH",
+                shape=(self.npoints, self.stratStep),
+                dtype="float64",
                 compression="gzip",
             )
-            f["erodep"][:, 0] = self.cumEDLocal.getArray()
+            f["stratH"][:, : self.stratStep] = self.stratH[:, : self.stratStep]
+
+            # Write stratal layers fine percentage per layers
+            f.create_dataset(
+                "stratF",
+                shape=(self.npoints, self.stratStep),
+                dtype="float64",
+                compression="gzip",
+            )
+            f["stratF"][:, : self.stratStep] = self.stratF[:, : self.stratStep]
+
+            # Write porosity values for coarse sediments
+            f.create_dataset(
+                "phiS",
+                shape=(self.npoints, self.stratStep),
+                dtype="float64",
+                compression="gzip",
+            )
+            f["phiS"][:, : self.stratStep] = self.phiS[:, : self.stratStep]
+
+            # Write porosity values for fine sediments
+            f.create_dataset(
+                "phiF",
+                shape=(self.npoints, self.stratStep),
+                dtype="float64",
+                compression="gzip",
+            )
+            f["phiF"][:, : self.stratStep] = self.phiF[:, : self.stratStep]
+
+            # Write stratal layers carbonate percentage per layers
+            if self.carbOn:
+                f.create_dataset(
+                    "stratC",
+                    shape=(self.npoints, self.stratStep),
+                    dtype="float64",
+                    compression="gzip",
+                )
+                f["stratC"][:, : self.stratStep] = self.stratC[:, : self.stratStep]
+
+                # Write porosity values for carbonate sediments
+                f.create_dataset(
+                    "phiC",
+                    shape=(self.npoints, self.stratStep),
+                    dtype="float64",
+                    compression="gzip",
+                )
+                f["phiC"][:, : self.stratStep] = self.phiC[:, : self.stratStep]
 
         MPIcomm.Barrier()
+
         if MPIrank == 0 and self.verbose:
             print(
                 "Creating stratal outputfile \
@@ -147,14 +270,25 @@ class WriteMesh(object):
                 flush=True,
             )
 
-        self.stratStep += 1
-
         return
 
     def _outputMesh(self):
         """
-        Saves mesh local information stored in the DMPlex to HDF5 file
-        If the file already exists, it is overwritten.
+        Saves mesh local information stored in the DMPlex to HDF5 file. If the file already
+        exists, it will be overwritten. Mesh characteristics are recorded for each partition.
+        The following variables will be available:
+
+        - surface elevation `elev`.
+        - cumulative erosion & deposition values `erodep`.
+        - flow accumulation `flowAcc` before pit filling.
+        - flow accumulation `fillAcc` for depressionless surface.
+        - river sediment load `sedLoad`.
+        - fine sediment load `sedLoadf` when dual lithologies are accounted for.
+        - carbonate sediment load `sedLoadc` when carbonate module is turned on.
+        - uplift subsidence values if vertical tectonic forcing is considered `uplift`.
+        - horizontal displacement values when considered `hdisp`.
+        - precipitation maps based on forcing conditions `rain`.
+
         """
 
         t = process_time()
@@ -163,7 +297,7 @@ class WriteMesh(object):
             with h5py.File(topology, "w") as f:
                 f.create_dataset(
                     "coords",
-                    shape=(len(self.lcoords[:, 0]), 3),
+                    shape=(self.npoints, 3),
                     dtype="float32",
                     compression="gzip",
                 )
@@ -176,7 +310,7 @@ class WriteMesh(object):
                 )
                 f["cells"][:, :] = self.lcells + 1
             self.elems = MPIcomm.gather(len(self.lcells[:, 0]), root=0)
-            self.nodes = MPIcomm.gather(len(self.lcoords[:, 0]), root=0)
+            self.nodes = MPIcomm.gather(self.npoints, root=0)
 
         h5file = (
             self.outputDir
@@ -189,41 +323,66 @@ class WriteMesh(object):
             + ".h5"
         )
         with h5py.File(h5file, "w") as f:
-            if self.stratStep == 0:
-                f.create_dataset(
-                    "elev",
-                    shape=(len(self.lcoords[:, 0]), 1),
-                    dtype="float32",
-                    compression="gzip",
-                )
-                f["elev"][:, 0] = self.hLocal.getArray()
-                f.create_dataset(
-                    "erodep",
-                    shape=(len(self.lcoords[:, 0]), 1),
-                    dtype="float32",
-                    compression="gzip",
-                )
-                f["erodep"][:, 0] = self.cumEDLocal.getArray()
+            # if self.stratStep == 0:
+            f.create_dataset(
+                "elev",
+                shape=(self.npoints, 1),
+                dtype="float32",
+                compression="gzip",
+            )
+            f["elev"][:, 0] = self.hLocal.getArray()
+            f.create_dataset(
+                "erodep",
+                shape=(self.npoints, 1),
+                dtype="float32",
+                compression="gzip",
+            )
+            f["erodep"][:, 0] = self.cumEDLocal.getArray()
             f.create_dataset(
                 "flowAcc",
-                shape=(len(self.lcoords[:, 0]), 1),
+                shape=(self.npoints, 1),
                 dtype="float32",
                 compression="gzip",
             )
             data = self.FAL.getArray().copy()
             data[data <= 0.0] = 1.0
             f["flowAcc"][:, 0] = data
+            f.create_dataset(
+                "fillAcc",
+                shape=(self.npoints, 1),
+                dtype="float32",
+                compression="gzip",
+            )
+            data = self.fillFAL.getArray().copy()
+            data[data <= 0.0] = 1.0
+            f["fillAcc"][:, 0] = data
             f.create_dataset(
                 "sedLoad",
-                shape=(len(self.lcoords[:, 0]), 1),
+                shape=(self.npoints, 1),
                 dtype="float32",
                 compression="gzip",
             )
             f["sedLoad"][:, 0] = self.vSedLocal.getArray().copy()
+            if self.stratNb > 0:
+                f.create_dataset(
+                    "sedLoadf",
+                    shape=(self.npoints, 1),
+                    dtype="float32",
+                    compression="gzip",
+                )
+                f["sedLoadf"][:, 0] = self.vSedfLocal.getArray().copy()
+                if self.carbOn:
+                    f.create_dataset(
+                        "sedLoadc",
+                        shape=(self.npoints, 1),
+                        dtype="float32",
+                        compression="gzip",
+                    )
+                    f["sedLoadc"][:, 0] = self.vSedcLocal.getArray().copy()
             if self.uplift is not None:
                 f.create_dataset(
                     "uplift",
-                    shape=(len(self.lcoords[:, 0]), 1),
+                    shape=(self.npoints, 1),
                     dtype="float32",
                     compression="gzip",
                 )
@@ -231,7 +390,7 @@ class WriteMesh(object):
             if self.hdisp is not None:
                 f.create_dataset(
                     "hdisp",
-                    shape=(len(self.lcoords[:, 0]), 3),
+                    shape=(self.npoints, 3),
                     dtype="float32",
                     compression="gzip",
                 )
@@ -239,7 +398,7 @@ class WriteMesh(object):
             if self.rainVal is not None:
                 f.create_dataset(
                     "rain",
-                    shape=(len(self.lcoords[:, 0]), 1),
+                    shape=(self.npoints, 1),
                     dtype="float32",
                     compression="gzip",
                 )
@@ -247,6 +406,9 @@ class WriteMesh(object):
 
             del data
 
+        if self.stratNb > 0 and self.stratStep > 0:
+            self._outputStrat()
+
         if MPIrank == 0:
             self._save_DMPlex_XMF()
             self._save_XDMF()
@@ -268,44 +430,62 @@ class WriteMesh(object):
 
     def readData(self):
         """
-        For restarted simulations, this function reads the previous dataset.
+        When a simulation restarts, variables from previous HDF5 output files are read and assigned
+        to the restarting run.
+
+        The following variables are used:
+
+        - surface elevation `elev`.
+        - cumulative erosion & deposition values `erodep`.
+        - flow accumulation `flowAcc` before pit filling.
+        - flow accumulation `fillAcc` for depressionless surface.
+        - river sediment load `sedLoad`.
+        - fine sediment load `sedLoadf` when dual lithologies are accounted for.
+        - carbonate sediment load `sedLoadc` when carbonate module is turned on.
+
         """
 
-        if self.stratStep == 0:
-            h5file = (
-                self.outputDir
-                + "/h5/"
-                + self.file
-                + "."
-                + str(self.step)
-                + ".p"
-                + str(MPIrank)
-                + ".h5"
-            )
-            if os.path.exists(h5file):
-                hf = h5py.File(h5file, "r")
-            else:
-                raise ValueError("Restart file is missing...")
-
-            self.hLocal.setArray(np.array(hf["/elev"])[:, 0])
-            self.dm.localToGlobal(self.hLocal, self.hGlobal)
-            self.cumEDLocal.setArray(np.array(hf["/erodep"])[:, 0])
-            self.dm.localToGlobal(self.cumEDLocal, self.cumED)
-            self.vSedLocal.setArray(np.array(hf["/sedLoad"])[:, 0])
-            self.dm.localToGlobal(self.vSedLocal, self.vSed)
-            self.FAL.setArray(np.array(hf["/flowAcc"])[:, 0])
-            self.dm.localToGlobal(self.FAL, self.FAG)
-
-            self.elems = MPIcomm.gather(len(self.lcells[:, 0]), root=0)
-            self.nodes = MPIcomm.gather(len(self.lcoords[:, 0]), root=0)
-
-            hf.close()
-
+        # if self.stratStep == 0:
+        h5file = (
+            self.outputDir
+            + "/h5/"
+            + self.file
+            + "."
+            + str(self.step - 1)
+            + ".p"
+            + str(MPIrank)
+            + ".h5"
+        )
+        if os.path.exists(h5file):
+            hf = h5py.File(h5file, "r")
         else:
+            raise ValueError("Restart file is missing...")
+        self.hLocal.setArray(np.array(hf["/elev"])[:, 0])
+        self.dm.localToGlobal(self.hLocal, self.hGlobal)
+        self.cumEDLocal.setArray(np.array(hf["/erodep"])[:, 0])
+        self.dm.localToGlobal(self.cumEDLocal, self.cumED)
+        self.vSedLocal.setArray(np.array(hf["/sedLoad"])[:, 0])
+        self.dm.localToGlobal(self.vSedLocal, self.vSed)
+        self.FAL.setArray(np.array(hf["/flowAcc"])[:, 0])
+        self.dm.localToGlobal(self.FAL, self.FAG)
+        self.fillFAL.setArray(np.array(hf["/fillAcc"])[:, 0])
+        self.dm.localToGlobal(self.fillFAL, self.fillFAG)
+        if self.stratNb > 0:
+            self.vSedfLocal.setArray(np.array(hf["/sedLoadf"])[:, 0])
+            self.dm.localToGlobal(self.vSedfLocal, self.vSedf)
+            if self.carbOn:
+                self.vSedcLocal.setArray(np.array(hf["/sedLoadc"])[:, 0])
+                self.dm.localToGlobal(self.vSedcLocal, self.vSedc)
+        self.elems = MPIcomm.gather(len(self.lcells[:, 0]), root=0)
+        self.nodes = MPIcomm.gather(len(self.lcoords[:, 0]), root=0)
+
+        hf.close()
+
+        if self.stratNb > 0 and self.stratStep > 0:
             h5file = (
                 self.outputDir
                 + "/h5/stratal."
-                + str(self.stratStep - 1)
+                + str(self.step - 1)
                 + ".p"
                 + str(MPIrank)
                 + ".h5"
@@ -315,35 +495,13 @@ class WriteMesh(object):
             else:
                 raise ValueError("Restart file is missing...")
 
-            self.hLocal.setArray(np.array(hf["/elev"])[:, 0])
-            self.dm.localToGlobal(self.hLocal, self.hGlobal)
-            self.cumEDLocal.setArray(np.array(hf["/erodep"])[:, 0])
-            self.dm.localToGlobal(self.cumEDLocal, self.cumED)
-
-            self.elems = MPIcomm.gather(len(self.lcells[:, 0]), root=0)
-            self.nodes = MPIcomm.gather(len(self.lcoords[:, 0]), root=0)
-
-            hf.close()
-
-            h5file = (
-                self.outputDir
-                + "/h5/"
-                + self.file
-                + "."
-                + str(self.step)
-                + ".p"
-                + str(MPIrank)
-                + ".h5"
-            )
-            if os.path.exists(h5file):
-                hf = h5py.File(h5file, "r")
-            else:
-                raise ValueError("Restart file is missing...")
-
-            self.vSedLocal.setArray(np.array(hf["/sedLoad"])[:, 0])
-            self.dm.localToGlobal(self.vSedLocal, self.vSed)
-            self.FAL.setArray(np.array(hf["/flowAcc"])[:, 0])
-            self.dm.localToGlobal(self.FAL, self.FAG)
+            self.stratZ.fill(0.0)
+            self.stratZ[:, : self.stratStep] = np.array(hf["/stratZ"])
+            self.stratH.fill(0.0)
+            self.stratH[:, : self.stratStep] = np.array(hf["/stratH"])
+            if self.carbOn:
+                self.stratC.fill(0.0)
+                self.stratC[:, : self.stratStep] = np.array(hf["/stratC"])
 
             hf.close()
 
@@ -352,22 +510,35 @@ class WriteMesh(object):
     def _forcePaleo(self):
         """
         Forcing the model based on backward simulation results. This function computes
-        difference between backward model elevation and current one
+        difference between backward model elevation and current one.
+
+        .. note::
+
+            The backward elevation is often computed based on paleo-elevation maps at specific
+            time interval. During the pre-processing phase a series of backward elevations are
+            then created based on these paleo-elevation maps. These backward elevations are then
+            used to force the forward model over time. This is done by computing the differences
+            at any given time step between the obtained forward elevations and the predicted backward
+            ones. These differences are then used to define an uplift map that matches the differences
+            modulated by an imposed factor (`alpha`). Once the uplift map has been computed the model
+            is then rerun from the previous time step to the current one.
+
         """
 
         # Store the vertical displacement in the uplift variable
         self._upliftBF(self.alpha[self.forceStep], self.stepb[self.forceStep])
 
         # Update model parameters to the former tectonic time step
-        self.tNow -= self.tecStep
+        self.tNow -= self.tout
         self.step -= 1
-
+        n = int(self.tout / self.tecStep)
+        self.tecNb -= n
         if self.strat > 0:
             n = int(self.tout / self.strat)
             self.stratStep = (self.step - 1) * n
         self.saveTime = self.tNow + self.tout
 
-        if self.saveStrat <= self.tEnd:
+        if self.saveStrat <= self.tEnd + self.strat:
             self.saveStrat = self.tNow + self.strat
 
         # Getting PETSc vectors values
@@ -378,6 +549,12 @@ class WriteMesh(object):
             self.dm.localToGlobal(self.hLocal, self.hGlobal)
             self.vSed.set(0.0)
             self.vSedLocal.set(0.0)
+            if self.stratNb > 0:
+                self.vSedf.set(0.0)
+                self.vSedfLocal.set(0.0)
+                if self.carbOn:
+                    self.vSedc.set(0.0)
+                    self.vSedcLocal.set(0.0)
             self.cumED.set(0.0)
             self.cumEDLocal.set(0.0)
         else:
@@ -401,16 +578,82 @@ class WriteMesh(object):
             print("Backward file: {} is missing!".format(h5file), flush=True)
             raise ValueError("Backward file is missing...")
 
-        self.uplift = np.array(hf["/elev"])[:, 0] - self.hLocal.getArray()
+        # Get the difference between backward and forward model on TIN
+        ldiff = np.array(hf["/elev"])[:, 0] - self.hLocal.getArray()
+
+        # erosional features using a low-pass filter
+        gdiff = ldiff[self.lgIDs]
+        if MPIsize > 1:
+            temp = np.full(self.shadowgNb, -1.0e8, dtype=np.float64)
+            temp[self.gshadinIDs] = gdiff[self.gshadowIDs]
+            temp[self.gshadoutIDs] = -1.0e8
+            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
+            gdiff[self.shadowAlls] = temp
+        ldiff = gdiff[self.glIDs]
+
+        # self.uplift = ldiff
+        # self.uplift *= alpha / self.tecStep
+        # hf.close()
+        # return
+
+        # Map it on a regularly spaced mesh (lon/lat)
+        diffreg = (
+            np.sum(self.forceWeights * ldiff[self.forceIDs], axis=1) * self.forceTop
+        )
+        if len(self.forceonIDs) > 0:
+            diffreg[self.forceonIDs] = ldiff[self.forceIDs[self.forceonIDs, 0]]
+
+        # Apply a low-pass filter to the regular array
+        # median = False
+        kernel_size = 3
+        # if median:
+        #     filter = ndimage.median_filter(
+        #         diffreg.reshape(self.regshape), size=kernel_size, mode="wrap"
+        #     )
+        # else:
+        filter = ndimage.gaussian_filter(
+            diffreg.reshape(self.regshape), sigma=kernel_size, mode="wrap"
+        )
+        filter = filter.flatten()
+
+        # Map it back to the spherical mesh
+        lfilter = np.sum(self.regWeights * filter[self.regIDs], axis=1) * self.regTop
+        if len(self.regonIDs) > 0:
+            lfilter[self.regonIDs] = filter[self.regIDs[self.regonIDs, 0]]
+
+        # From local to global
+        gfilter = lfilter[self.lgIDs]
+        if MPIsize > 1:
+            temp = np.full(self.shadowgNb, -1.0e8, dtype=np.float64)
+            temp[self.gshadinIDs] = gfilter[self.gshadowIDs]
+            temp[self.gshadoutIDs] = -1.0e8
+            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
+            gfilter[self.shadowAlls] = temp
+
+        # Specify new uplift value matching backward elevation removing the
+        # erosional features using a low-pass filter
+        self.uplift = gfilter[self.glIDs]
+        self.uplift[self.seaID] = ldiff[self.seaID]
+
         self.uplift *= alpha / self.tecStep
         hf.close()
 
+        del diffreg, filter, lfilter, gfilter, ldiff, hf
+        if MPIsize > 1:
+            del temp
+        gc.collect()
+
         return
 
     def _save_DMPlex_XMF(self):
         """
-        Saves mesh local information stored in the HDF5 to XMF file
-        to visualise in Paraview.
+        Saves mesh local information stored in the HDF5 to XmF file. The XMF files are XML schema
+        explaining how to read `gospl` data files.
+
+        The XmF file is written by a single processor (rank 0) and contains each partition HDF5 files
+        in blocks. The variables described for the HDF5 file (function `_outputMesh` above) are all
+        accessible from this file.
+
         """
 
         xmf_file = self.outputDir + "/xmf/" + self.file + str(self.step) + ".xmf"
@@ -429,18 +672,6 @@ class WriteMesh(object):
             pfile = (
                 "h5/" + str(self.file) + "." + str(self.step) + ".p" + str(p) + ".h5"
             )
-            if self.stratStep == 0:
-                afile = (
-                    "h5/"
-                    + str(self.file)
-                    + "."
-                    + str(self.step)
-                    + ".p"
-                    + str(p)
-                    + ".h5"
-                )
-            else:
-                afile = "h5/stratal." + str(self.stratStep - 1) + ".p" + str(p) + ".h5"
             tfile = "h5/topology.p" + str(p) + ".h5"
             f.write('      <Grid Name="Block.%s">\n' % (str(p)))
             f.write(
@@ -464,7 +695,7 @@ class WriteMesh(object):
             f.write(
                 '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
             )
-            f.write('Dimensions="%d 1">%s:/elev</DataItem>\n' % (self.nodes[p], afile))
+            f.write('Dimensions="%d 1">%s:/elev</DataItem>\n' % (self.nodes[p], pfile))
             f.write("         </Attribute>\n")
 
             f.write('         <Attribute Type="Scalar" Center="Node" Name="ED">\n')
@@ -472,7 +703,7 @@ class WriteMesh(object):
                 '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
             )
             f.write(
-                'Dimensions="%d 1">%s:/erodep</DataItem>\n' % (self.nodes[p], afile)
+                'Dimensions="%d 1">%s:/erodep</DataItem>\n' % (self.nodes[p], pfile)
             )
             f.write("         </Attribute>\n")
 
@@ -485,6 +716,15 @@ class WriteMesh(object):
             )
             f.write("         </Attribute>\n")
 
+            f.write('         <Attribute Type="Scalar" Center="Node" Name="fillFA">\n')
+            f.write(
+                '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
+            )
+            f.write(
+                'Dimensions="%d 1">%s:/fillAcc</DataItem>\n' % (self.nodes[p], pfile)
+            )
+            f.write("         </Attribute>\n")
+
             f.write('         <Attribute Type="Scalar" Center="Node" Name="SL">\n')
             f.write(
                 '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
@@ -493,6 +733,29 @@ class WriteMesh(object):
                 'Dimensions="%d 1">%s:/sedLoad</DataItem>\n' % (self.nodes[p], pfile)
             )
             f.write("         </Attribute>\n")
+            if self.stratNb > 0:
+                f.write('         <Attribute Type="Scalar" Center="Node" Name="SLf">\n')
+                f.write(
+                    '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
+                )
+                f.write(
+                    'Dimensions="%d 1">%s:/sedLoadf</DataItem>\n'
+                    % (self.nodes[p], pfile)
+                )
+                f.write("         </Attribute>\n")
+                if self.carbOn:
+                    f.write(
+                        '         <Attribute Type="Scalar" Center="Node" Name="SLc">\n'
+                    )
+                    f.write(
+                        '          <DataItem Format="HDF" NumberType="Float" Precision="4" '
+                    )
+                    f.write(
+                        'Dimensions="%d 1">%s:/sedLoadc</DataItem>\n'
+                        % (self.nodes[p], pfile)
+                    )
+                    f.write("         </Attribute>\n")
+
             if self.hdisp is not None:
                 f.write(
                     '         <Attribute Type="Vector" Center="Node" Name="hTec">\n'
@@ -552,7 +815,14 @@ class WriteMesh(object):
 
     def _save_XDMF(self):
         """
-        This function writes the XDmF file which is calling the XmF file.
+        This function writes the XDmF file which is calling the XmF files above. The XDmF file
+        represents the *time series* of the model outputs and can be directly loaded and visualised
+        with `Paraview <https://www.paraview.org/download/>`_.
+
+        .. note::
+
+            For a brief overview of the approach used to record `gospl` outputs, user can read
+            this `visit documentation <https://www.visitusers.org/index.php?title=Using_XDMF_to_read_HDF5>`_
         """
 
         xdmf_file = self.outputDir + "/" + self.file + ".xdmf"
